#
# Copyright (c) Elliot Peele <elliot@bentlogic.net>
#
# This program is distributed under the terms of the MIT License as found
# in a file called LICENSE. If it is not present, the license
# is always available at http://www.opensource.org/licenses/mit-license.php.
#
# This program is distributed in the hope that it will be useful, but
# without any warrenty; without even the implied warranty of merchantability
# or fitness for a particular purpose. See the MIT License for full details.
#

"""
RESTful client library, meant to be used with REST APIs generated by prism_rest
and possibly others.
"""

from .client import Client
from .cache import InstanceCache

def open(uri, headers=None, verify=False):
    return InstanceCache(Client(headers=headers, verify=verify)).get(uri)



#
# Copyright (c) Elliot Peele <elliot@bentlogic.net>
#
# This program is distributed under the terms of the MIT License as found
# in a file called LICENSE. If it is not present, the license
# is always available at http://www.opensource.org/licenses/mit-license.php.
#
# This program is distributed in the hope that it will be useful, but
# without any warrenty; without even the implied warranty of merchantability
# or fitness for a particular purpose. See the MIT License for full details.
#

from .resource import BaseResource

class Collection(BaseResource):
    def __getitem__(self, idx):
        return self._data[idx]

    def __setitem__(self, idx, value):
        pass

    def __delitem__(self, idx):
        self._cache.delete(self._data[idx].delete())
        del self._data[idx]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __nonzero__(self):
        return bool(self._data)

    def append(self, resource, post=True):
        if isinstance(resource, BaseResource):
            value = resource._data
        else:
            value = resource

        if post:
            obj = self._cache.post(self._uri, value)
            self._data.append(obj)
            return obj

        else:
            self._data.append(resource)

    def refresh(self):
        resource = self._cache.get(self._uri, cache=False)
        assert isinstance(resource, Collection)
        self._data = resource._data
        self._metadata = resource._metadata



#
# Copyright (c) Elliot Peele <elliot@bentlogic.net>
#
# This program is distributed under the terms of the MIT License as found
# in a file called LICENSE. If it is not present, the license
# is always available at http://www.opensource.org/licenses/mit-license.php.
#
# This program is distributed in the hope that it will be useful, but
# without any warrenty; without even the implied warranty of merchantability
# or fitness for a particular purpose. See the MIT License for full details.
#

import json

from .renderer import JSONRenderer

class Default(object): pass

class InstanceCache(dict):
    __slots__ = ('client', 'resp_cache', )

    def __init__(self, client):
        self.client = client
        self.resp_cache = ResponseCache()

    def __getitem__(self, uri, force=False, cache=True, params=None):
        if uri in self and not force and cache:
            return dict.__getitem__(self, uri)
        else:
            resp = self.client.get(uri, params=params)
            return self.get_by_response(resp, cache=cache)

    def get(self, uri, default=Default, force=False, cache=True, params=None):
        if uri not in self and not force and default is not Default:
            return default
        else:
            return self.__getitem__(uri, force=force, cache=cache,
                    params=params)

    def get_by_response(self, resp, cache=True):
        resp.close()
        if resp.status_code >= 400:
            import epdb; epdb.st()
        resource = resp.json(object_hook=JSONRenderer(self))
#        self.resp_cache[resource._uri] = resp

        if cache:
            self[resource._uri] = resource
        return resource

    def delete(self, uri):
        del self[uri]
#        del self.resp_cache[uri]
        self.client.delete(uri).close()

    def post(self, uri, data):
        resp = self.client.post(uri, json.dumps(data))
        return self.get_by_response(resp, cache=True)

    def put(self, uri, data):
        resp = self.client.put(uri, json.dumps(data))
        resource = self.get_by_response(resp, cache=False)
        self[resource._uri]._data = resource._data


class ResponseCache(dict):
    __slots__ = ()



#
# Copyright (c) Elliot Peele <elliot@bentlogic.net>
#
# This program is distributed under the terms of the MIT License as found
# in a file called LICENSE. If it is not present, the license
# is always available at http://www.opensource.org/licenses/mit-license.php.
#
# This program is distributed in the hope that it will be useful, but
# without any warrenty; without even the implied warranty of merchantability
# or fitness for a particular purpose. See the MIT License for full details.
#

import logging

log = logging.getLogger('prism_rest_client.util')

class AttrDict(dict):
    """
    Dictionary implementation that lets you refer to dictionary keys as
    normal attributes.
    """

    __slots__ = ()

    def __getattr__(self, attr):
        if attr.startswith('_'):
            return dict.__getattribute__(self, attr)
        else:
            return self.__getitem__(attr)

    def __setattr__(self, attr, value):
        if attr.startswith('_'):
            return dict.__setattr__(self, attr, value)
        else:
            return self.__setitem__(attr, value)



#
# Copyright (c) Elliot Peele <elliot@bentlogic.net>
#
# This program is distributed under the terms of the MIT License as found
# in a file called LICENSE. If it is not present, the license
# is always available at http://www.opensource.org/licenses/mit-license.php.
#
# This program is distributed in the hope that it will be useful, but
# without any warrenty; without even the implied warranty of merchantability
# or fitness for a particular purpose. See the MIT License for full details.
#

from .lib.util import AttrDict
from .resource import Resource
from .collection import Collection

class JSONRenderer(object):
    def __init__(self, cache):
        self.cache = cache

    def __call__(self, pairs):
        uri = pairs.get('id')
        md = pairs.get('metadata')
        pairs = AttrDict(pairs)

        if not uri:
            return pairs

        data = pairs.get('data')
        if data is not None:
            resource = Collection(uri, self.cache, data=data, metadata=md)

        else:
            resource = Resource(uri, self.cache, data=pairs, metadata=md)

        return resource



#
# Copyright (c) Elliot Peele <elliot@bentlogic.net>
#
# This program is distributed under the terms of the MIT License as found
# in a file called LICENSE. If it is not present, the license
# is always available at http://www.opensource.org/licenses/mit-license.php.
#
# This program is distributed in the hope that it will be useful, but
# without any warrenty; without even the implied warranty of merchantability
# or fitness for a particular purpose. See the MIT License for full details.
#

import types

class BaseResource(object):
    def __init__(self, uri, cache, data, metadata):
        self._uri = uri
        self._cache = cache
        self._data = data
        self._metadata = metadata


class Resource(BaseResource):
    def __init__(self, uri, cache, data, metadata, parent=None):
        BaseResource.__init__(self, uri, cache, data, metadata)
        self._parent = parent
        self._dirty = False
        self._local_cache = {}

    def __getattr__(self, name, params=None):
        val = self._data.get(name)
        obj = self._getObj(name, val, params=params)
        if obj is not None:
            return obj
        elif val is None:
            raise AttributeError, '%s not in resource' % name
        else:
            return val

    def __setattr__(self, name, value):
        if name.startswith('_'):
            object.__setattr__(self, name, value)
            return

        if name in self._data:
            val = self._data.get(name)
            if not self._setObj(val, name, value):
                self._dirty = True
                self._data[name] = value
        else:
            self._dirty = True

            if isinstance(value, list) and len(value) == 0:
                value = self.__class__(self._uri, self._cache, )

            self._data[name] = value

    def _getObj(self, name, val, params=None):
        if isinstance(val, types.StringTypes) and (val.startswith('http://')
                or val.startswith('https://')):
            return self._cache.get(val, params=params)

        if isinstance(val, dict):
            if 'id' in val:
                return self._cache.get(val['id'])

            valId = id(val)
            if valId not in self._local_cache:
                self._local_cache[valId] = self.__class__(self._uri,
                        self._cache, data=val, metadata=self._metadata,
                        parent=self)
            return self._local_cache.get(valId)

    def _setObj(self, val, name, value):
        return False

    def query(self, _n, **kwargs):
        return self.__getattr__(_n, params=kwargs)

    def persist(self):
        if self._dirty:
            self._cache.put(self._uri, self._data)
            self._dirty = False

    def delete(self):
        self._cache.delete(self._uri)

    def refresh(self, force=False):
        if not self._dirty or force:
            self._dirty = False
            self._local_cache = {}

            resource = self._cache.get(self._uri, cache=False)
            self._data = resource._data
            self._metadata = resource._metadata



#
# Copyright (c) Elliot Peele <elliot@bentlogic.net>
#
# This program is distributed under the terms of the MIT License as found
# in a file called LICENSE. If it is not present, the license
# is always available at http://www.opensource.org/licenses/mit-license.php.
#
# This program is distributed in the hope that it will be useful, but
# without any warrenty; without even the implied warranty of merchantability
# or fitness for a particular purpose. See the MIT License for full details.
#

import requests

class Client(object):
    def __init__(self, headers=None, verify=True):
        self.headers = headers
        self.verify = verify

    def get(self, uri, params=None):
        return requests.get(uri, headers=self.headers,
                params=params and params or {},
                verify=self.verify)

    def post(self, uri, data):
        return requests.post(uri, data, headers=self.headers,
                verify=self.verify)

    def delete(self, uri):
        return requests.delete(uri, headers=self.headers, verify=self.verify)

    def put(self, uri, data):
        return requests.put(uri, data, headers=self.headers,
                verify=self.verify)



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()

requires = [
        'requests',
    ]

setup(name='prism_rest_client',
      version='0.2',
      description='prism_rest_client',
      long_description=README,
      classifiers=[
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        ],
      author='Elliot Peele',
      author_email='elliot@bentlogic.net',
      url='https://github.com/elliotpeele/prism_rest_client',
      keywords='web rest client',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='prism_rest_client',
      install_requires=requires,
      )
